#include "MockAnnouncer.h"
#include "FlightInformation.h"

using namespace testing;

const std::string hello("Hello, World");
const std::string goodbye("And so, farewell");

TEST(FlightInfoTest, canAnnounceWithoutSubscribing)
{
    FlightInformation info;
    EXPECT_NO_THROW(info.announce(hello));
}

class FlightInfo_Fixture : public Test
{
protected:
    void SetUp()
    {
        announcerId = info.subscribe(announcer);

        DefaultValue<bool>::Set(true);
        //ON_CALL(announcer, isOperational()).WillByDefault(Return(true));
        //EXPECT_CALL(announcer, isOperational()).WillOnce(Return(true));
    }

protected:
    FlightInformation info;

    NiceMock<MockAnnouncer> announcer;
    int announcerId;
};

TEST_F(FlightInfo_Fixture, announcementShouldBePassedToSubscriber)
{
    EXPECT_CALL(announcer, display(hello)).Times(1);

    EXPECT_NO_THROW(info.announce(hello));
}

TEST_F(FlightInfo_Fixture, allAnnouncementsShouldBePassedToSubscriber)
{
    EXPECT_CALL(announcer, display(hello)).Times(1);
    EXPECT_CALL(announcer, display(goodbye)).Times(1);

    EXPECT_NO_THROW(info.announce(hello));
    EXPECT_NO_THROW(info.announce(goodbye));
}

TEST_F(FlightInfo_Fixture, demonstrateExpectationBehaviours)
{
    // Ordering
    // Sticky & Saturation - RetireOnSaturation

    EXPECT_CALL(announcer,display(hello)).Times(1);
    EXPECT_CALL(announcer,display(_)).Times(1).RetiresOnSaturation();

    EXPECT_NO_THROW(info.announce(hello));
    EXPECT_NO_THROW(info.announce(goodbye));
}

TEST_F(FlightInfo_Fixture, announcementsShouldBePassedToSubscriberInCorrectOrder)
{
    {
        testing::Sequence s;
        EXPECT_CALL(announcer, display(hello)).Times(1);
        EXPECT_CALL(announcer, display(goodbye)).Times(1);
    }

    EXPECT_NO_THROW(info.announce(hello));
    EXPECT_NO_THROW(info.announce(goodbye));
}

TEST_F(FlightInfo_Fixture, announcementShouldBePassedToAllSubscribers)
{
    MockAnnouncer announcer2;
    info.subscribe(announcer2);

    EXPECT_CALL(announcer2,isOperational()).WillOnce(Return(true));
    EXPECT_CALL(announcer, display(hello)).Times(1);
    EXPECT_CALL(announcer2, display(hello)).Times(1);

    EXPECT_NO_THROW(info.announce(hello));
}

TEST_F(FlightInfo_Fixture, announcementShouldNotBePassedOnceUnsubscribed)
{
    EXPECT_CALL(announcer, display(_)).Times(1);

    EXPECT_NO_THROW(info.announce(hello));
    EXPECT_NO_THROW(info.unsubscribe(announcerId));
    EXPECT_NO_THROW(info.announce(hello));
}

TEST_F(FlightInfo_Fixture, announcementsShouldNotBeSentToNonOperationalSubscribers)
{
    EXPECT_CALL(announcer, isOperational()).WillOnce(Return(false));

    EXPECT_NO_THROW(info.announce(hello));
}

TEST_F(FlightInfo_Fixture, announcementsShouldBeSentOnceSubscriberBecomesOperational)
{
    EXPECT_CALL(announcer, isOperational())
        .WillOnce(Return(false))
        .WillRepeatedly(Return(true));
    EXPECT_CALL(announcer, display(_)).Times(2);

    EXPECT_NO_THROW(info.announce(hello));
    EXPECT_NO_THROW(info.announce(hello));
    EXPECT_NO_THROW(info.announce(hello));
}

TEST_F(FlightInfo_Fixture, shouldAlwaysCheckIsOperationalBeforeDisplayingAnnouncement)
{
    MockAnnouncer announcer2;
    info.subscribe(announcer2);

    Sequence seqAnnouncer;
    Sequence seqAnnouncer2;

    EXPECT_CALL(announcer,isOperational()).InSequence(seqAnnouncer).WillOnce(Return(true));
    EXPECT_CALL(announcer, display(hello)).Times(1).InSequence(seqAnnouncer);

    EXPECT_CALL(announcer2,isOperational()).InSequence(seqAnnouncer2).WillOnce(Return(true));
    EXPECT_CALL(announcer2, display(hello)).Times(1).InSequence(seqAnnouncer, seqAnnouncer2);

    EXPECT_NO_THROW(info.announce(hello));
}

TEST_F(FlightInfo_Fixture, shouldAlwaysCheckIsOperationalBeforeDisplayingAnnouncement2)
{
    MockAnnouncer announcer2;
    info.subscribe(announcer2);

    Sequence seqAnnouncer;
    Sequence seqAnnouncer2;

    EXPECT_CALL(announcer,isOperational()).InSequence(seqAnnouncer).WillOnce(Return(true));
    Expectation expectation = EXPECT_CALL(announcer, display(hello)).Times(1).InSequence(seqAnnouncer);

    EXPECT_CALL(announcer2,isOperational()).InSequence(seqAnnouncer2).WillOnce(Return(true));
    EXPECT_CALL(announcer2, display(hello)).Times(1).InSequence(seqAnnouncer2).After(expectation);

    EXPECT_NO_THROW(info.announce(hello));
}


TEST_F(FlightInfo_Fixture, shouldAlwaysCheckIsOperationalBeforeDisplayingAnnouncement3)
{
    MockAnnouncer announcer2;
    info.subscribe(announcer2);

    ExpectationSet expectations;

    expectations += EXPECT_CALL(announcer,isOperational()).After(expectations).WillOnce(Return(true));
    expectations += EXPECT_CALL(announcer, display(hello)).Times(1).After(expectations);

    expectations += EXPECT_CALL(announcer2,isOperational()).After(expectations).WillOnce(Return(true));
    expectations += EXPECT_CALL(announcer2, display(hello)).Times(1).After(expectations);

    EXPECT_NO_THROW(info.announce(hello));
}
