#include "gmock/gmock.h"
#include <map>
#include <string>

using namespace testing;

class MyInterface
{
public:
    virtual void intMethod(int, int, int) =0;
    
    virtual void complexMethod(int, std::string, bool) =0;
};

class MyMock : public MyInterface
{
public:
    MOCK_METHOD3(intMethod, void(int, int, int));

    MOCK_METHOD3(complexMethod, void(int, std::string, bool));
};

class UnderTest
{
public:
    UnderTest(MyInterface& myInterface_)
        : myInterface(myInterface_)
    {
    }

    void doSomething(int i, int j, int k)
    {
        myInterface.intMethod(i, j, k);
    }

    void doSomething(int i, std::string s, bool b)
    {
        myInterface.complexMethod(i, s, b);
    }

private:
    MyInterface& myInterface;
};

class MultiTestsFixture : public Test
{
public:
    MultiTestsFixture()
        : underTest(mock)
    {
    }

protected:
    MyMock mock;
    UnderTest underTest;
};

TEST_F(MultiTestsFixture, shouldMatchOnSpecifiedArgs)
{
    EXPECT_CALL(mock, intMethod(_, _, _))
        .With(Args<0,1>(Lt()));

    EXPECT_NO_THROW(underTest.doSomething(1,2,3));
}

TEST_F(MultiTestsFixture, shouldMatchOnCompositionOfSpecifiedArgs)
{
    EXPECT_CALL(mock, intMethod(_, _, _))
        .With(AllOf(Args<0,1>(Lt()), Args<1,2>(Lt())));

    EXPECT_NO_THROW(underTest.doSomething(1,2,3));
}

// User defined
MATCHER(IsEven, "is even")
{
    return (arg%2) == 0;
}

TEST_F(MultiTestsFixture, shouldMatchOnUserDefinedMatcher)
{
    EXPECT_CALL(mock, intMethod(_, IsEven(), _));

    EXPECT_NO_THROW(underTest.doSomething(1,2,3));
}

// Parameterised User Defined
MATCHER_P2(IsInRange, low, high, "is within range %(low)s, %(high)s")
{
    return (arg>=low) && (arg<=high);
}

TEST_F(MultiTestsFixture, shouldMatchOnUserDefinedParameterisedMatcher)
{
    EXPECT_CALL(mock, intMethod(_, IsInRange(2,8), _));

    EXPECT_NO_THROW(underTest.doSomething(1,2,3));
}

// Composites
TEST_F(MultiTestsFixture, shouldMatchOnCompositeMatcher)
{
    EXPECT_CALL(mock, intMethod(_, AllOf(IsInRange(2,8), IsEven(), Gt(0)), _));

    EXPECT_NO_THROW(underTest.doSomething(1,2,3));
}

