#define GMOCK_HAS_REGEX 1

#include "gmock/gmock.h"

#include <string>

using namespace testing;

class MyInterface
{
public:
    virtual void unaryMethod(int) =0;

    virtual void overloadedMethod(long) =0;
    virtual void overloadedMethod(int) =0;
    virtual void overloadedMethod(float*) =0;
    virtual void overloadedMethod(double*) =0;

    virtual void cstringMethod(const char*) =0;
    virtual void stringMethod(const std::string&) =0;

    virtual void containerMethod(std::vector<int>&) =0;
};

class MyMock : public MyInterface
{
public:
    MOCK_METHOD1(unaryMethod, void(int));

    MOCK_METHOD1(overloadedMethod, void(int));
    MOCK_METHOD1(overloadedMethod, void(long));
    MOCK_METHOD1(overloadedMethod, void(float*));
    MOCK_METHOD1(overloadedMethod, void(double*));

    MOCK_METHOD1(cstringMethod, void(const char*));
    MOCK_METHOD1(stringMethod, void(const std::string&));

    MOCK_METHOD1(containerMethod, void(std::vector<int>&));
};

class UnderTest
{
public:
    UnderTest(MyInterface& myInterface_)
        : myInterface(myInterface_)
    {
    }

    void doSomething(int i)
    {
        myInterface.unaryMethod(i);
    }

    void doSomething(long l)
    {
        myInterface.overloadedMethod(l);
    }

    void doSomething(float* f)
    {
        myInterface.overloadedMethod(f);
    }

    void doSomething(double* d)
    {
        myInterface.overloadedMethod(d);
    }

    void doSomething(const char* s)
    {
        myInterface.cstringMethod(s);
    }

    void doSomething(const std::string& s)
    {
        myInterface.stringMethod(s);
    }

    void doSomething(std::vector<int>& c)
    {
        myInterface.containerMethod(c);
    }

private:
    MyInterface& myInterface;
};


class BasicMatcherFixture : public Test
{
public:
    BasicMatcherFixture()
        : underTest(mock)
    {
    }

protected:
    MyMock mock;
    UnderTest underTest;
};

TEST_F(BasicMatcherFixture, shouldMatchWildcard)
{
    EXPECT_CALL(mock, unaryMethod(_)).Times(3);

    EXPECT_NO_THROW(underTest.doSomething(1));
    EXPECT_NO_THROW(underTest.doSomething(2));
    EXPECT_NO_THROW(underTest.doSomething(3));
}

TEST_F(BasicMatcherFixture, shouldMatchTypeWildcard)
{
    EXPECT_CALL(mock, overloadedMethod(A<float*>())).Times(1);
    EXPECT_CALL(mock, overloadedMethod(A<double*>())).Times(1);

    float f;
    EXPECT_NO_THROW(underTest.doSomething(&f));

    double d;
    EXPECT_NO_THROW(underTest.doSomething(&d));
}

// String matchers
TEST_F(BasicMatcherFixture, shouldMatchOnCString)
{
    std::string statement("Hello");

    EXPECT_CALL(mock, cstringMethod(StrEq(statement.c_str()))).Times(1);

    EXPECT_NO_THROW(underTest.doSomething("Hello"));

    // Also StrCaseEq, StrCaseNe, StrNe
}

TEST_F(BasicMatcherFixture, shouldMatchOnPrefixAndSuffix)
{
    std::string prefix("Hello");
    std::string suffix("World");
    std::string statement = prefix + ", " + suffix;

    EXPECT_CALL(mock, stringMethod(StartsWith(prefix))).Times(1);
    EXPECT_CALL(mock, cstringMethod(EndsWith(suffix))).Times(1);

    EXPECT_NO_THROW(underTest.doSomething(statement));
    EXPECT_NO_THROW(underTest.doSomething(statement.c_str()));

    // Also HasSubstr()
}

TEST_F(BasicMatcherFixture, shouldMatchRegex)
{
    EXPECT_CALL(mock, cstringMethod(ContainsRegex(", "))).Times(1);
    EXPECT_CALL(mock, cstringMethod(MatchesRegex("^H.*d$"))).Times(1);

    EXPECT_NO_THROW(underTest.doSomething("Hello, World"));
    EXPECT_NO_THROW(underTest.doSomething("Heavens, Jeeves"));
}


// Containers
TEST_F(BasicMatcherFixture, shouldFindElementInContainer)
{
    int data[] = {0, 1, 2, 3, 4};
    std::vector<int> container(data, data+5);

    EXPECT_CALL(mock, containerMethod(Contains(4))).Times(1);
    EXPECT_CALL(mock, containerMethod(Contains(Lt(2)))).Times(1).RetiresOnSaturation();
    EXPECT_CALL(mock, containerMethod(ElementsAre(0,1,2,3,4))).Times(1).RetiresOnSaturation();
    EXPECT_CALL(mock, containerMethod(ElementsAreArray(data))).Times(1).RetiresOnSaturation();

    EXPECT_NO_THROW(underTest.doSomething(container));
    EXPECT_NO_THROW(underTest.doSomething(container));
    EXPECT_NO_THROW(underTest.doSomething(container));
    EXPECT_NO_THROW(underTest.doSomething(container));
}
// Also Le, Lt, Gt, Ge, Eq, Ne

// Generic
TEST_F(BasicMatcherFixture, shouldMatchByReference)
{
    std::string statement("Hello");
//    std::string statement2("Hello");

    // No copy made (clearly)
    EXPECT_CALL(mock, stringMethod(Ref(statement))).Times(1);

    EXPECT_NO_THROW(underTest.doSomething(statement));
//    EXPECT_NO_THROW(underTest.doSomething(statement2));
}

TEST_F(BasicMatcherFixture, shouldMatchByNullness)
{
    const char* null = NULL;
    const char* statement = "Hello";

    EXPECT_CALL(mock, cstringMethod(IsNull())).Times(1);
    EXPECT_CALL(mock, cstringMethod(NotNull())).Times(1);

    EXPECT_NO_THROW(underTest.doSomething(null));
    EXPECT_NO_THROW(underTest.doSomething(statement));
}

TEST_F(BasicMatcherFixture, shouldMatchCorrectOverloadedMethodWhenForced)
{
    EXPECT_CALL(mock, overloadedMethod(TypedEq<long>(3))).Times(1);
    //EXPECT_CALL(mock, overloadedMethod(3L)).Times(1);

    EXPECT_NO_THROW(underTest.doSomething(3L));
}

// Floating point
