#include "gmock/gmock.h"
#include <map>
#include <string>

using namespace testing;

struct SomeClass
{
    SomeClass(int i)
        : data(i)
    {
    }

    int getData() const
    {
        return data;
    }

    int data;
};

typedef std::map<std::string, int> MyCollection;
typedef std::pair<std::string, int> MyPair;

class MyInterface
{
public:
    virtual void udtRefMethod(SomeClass& o) =0;
    virtual void udtPtrMethod(SomeClass* o) =0;

    virtual void iteratorMethod(MyPair) =0;

    virtual void pointerMethod(int*) =0;

    virtual void intMethod(int) =0;
};

class MyMock : public MyInterface
{
public:
    MOCK_METHOD1(udtRefMethod, void(SomeClass&));
    MOCK_METHOD1(udtPtrMethod, void(SomeClass*));

    MOCK_METHOD1(iteratorMethod, void(MyPair));

    MOCK_METHOD1(pointerMethod, void(int*));

    MOCK_METHOD1(intMethod, void(int));
};

class UnderTest
{
public:
    UnderTest(MyInterface& myInterface_)
        : myInterface(myInterface_)
    {
    }

    void doSomething(SomeClass& obj)
    {
        myInterface.udtRefMethod(obj);
    }

    void doSomething(SomeClass* obj)
    {
        myInterface.udtPtrMethod(obj);
    }

    void doSomething(MyCollection& collection, int i)
    {
        MyCollection::iterator it;
        for (it=collection.begin(); it!=collection.end(); ++it)
        {
            if (it->second == i)
            {
                myInterface.iteratorMethod(*it);
            }
        }
    }

    void doSomething(int* p)
    {
        myInterface.pointerMethod(p);
    }

    void doSomething(int i)
    {
        myInterface.intMethod(i);
    }

private:
    MyInterface& myInterface;
};

class MemberTestsFixture : public Test
{
public:
    MemberTestsFixture()
        : underTest(mock)
    {
    }

protected:
    MyMock mock;
    UnderTest underTest;
};

TEST_F(MemberTestsFixture, shouldMatchFieldOfObjectViaRef)
{
    EXPECT_CALL(mock, udtRefMethod(Field(&SomeClass::data, Gt(0))));

    SomeClass object(5);
    EXPECT_NO_THROW(underTest.doSomething(object));
}

TEST_F(MemberTestsFixture, shouldMatchFieldOfObjectViaPtr)
{
    EXPECT_CALL(mock, udtPtrMethod(Field(&SomeClass::data, 5)));

    SomeClass object(5);
    EXPECT_NO_THROW(underTest.doSomething(&object));
}

TEST_F(MemberTestsFixture, shouldMatchPropertyOfObjectViaRef)
{
    EXPECT_CALL(mock, udtRefMethod(Property(&SomeClass::getData, Gt(0))));

    SomeClass object(5);
    EXPECT_NO_THROW(underTest.doSomething(object));
}

TEST_F(MemberTestsFixture, shouldMatchPropertyOfObjectViaPtr)
{
    EXPECT_CALL(mock, udtPtrMethod(Property(&SomeClass::getData, 5)));

    SomeClass object(5);
    EXPECT_NO_THROW(underTest.doSomething(&object));
}

TEST_F(MemberTestsFixture, shouldMatchIteratorKey)
{
    MyCollection collection;
    collection["One"] = 1;
    collection["Two"] = 2;
    collection["Three"] = 3;

    EXPECT_CALL(mock, iteratorMethod(Key("Two")));
    EXPECT_CALL(mock, iteratorMethod(Pair("Three", 3)));

    EXPECT_NO_THROW(underTest.doSomething(collection, 2));
    EXPECT_NO_THROW(underTest.doSomething(collection, 3));
}

// Pointer
TEST_F(MemberTestsFixture, shouldMatchUsingObjectPointedToByArgument)
{
    int i = 42;

    EXPECT_CALL(mock, pointerMethod(Pointee(i)));

    EXPECT_NO_THROW(underTest.doSomething(&i)); 
}

// Function
int divideByTwo(int p)
{
   return p / 2;
}

TEST_F(MemberTestsFixture, shouldMatchBasedOnReturnValueOfFunction)
{
    EXPECT_CALL(mock, intMethod(ResultOf(divideByTwo, 6)));

    EXPECT_NO_THROW(underTest.doSomething(12));
}

// Functor
class DivideByTwo : public std::unary_function<int, int>
{
public:
    int operator()(int p)
    {
       return p / 2;
    }
};

TEST_F(MemberTestsFixture, shouldMatchBasedOnReturnValueOfFunctor)
{
    EXPECT_CALL(mock, intMethod(ResultOf(DivideByTwo(), 3)));

    EXPECT_NO_THROW(underTest.doSomething(6));
}

// Predicate Adaptor
class IsEven : public std::unary_function<int, bool>
{
public:
    bool operator()(int p) const
    {
       return (p%2 == 0);
    }
};

TEST_F(MemberTestsFixture, shouldMatchBasedOnReturnValueOfPredicate)
{
    EXPECT_CALL(mock, intMethod(Truly(IsEven())));

    EXPECT_NO_THROW(underTest.doSomething(6));
}


