#include "gmock/gmock.h"

using namespace testing;

struct MyStruct
{
    MyStruct(const std::string& description_, int count_)
        : description(description_), count(count_)
    {
    }

    std::string description;
    int count;
};

class MyInterface
{
public:
    virtual int intMethod(int i, int j, int k) const =0;
    virtual MyStruct* newStructMethod() const =0;
    virtual const MyStruct& refStructMethod() const =0;
};

class MyMock : public MyInterface
{
public:
    MOCK_CONST_METHOD3(intMethod, int(int, int, int));
    MOCK_CONST_METHOD0(newStructMethod, MyStruct*());
    MOCK_CONST_METHOD0(refStructMethod, const MyStruct&());
};

class UnderTest
{
public:
    UnderTest(MyInterface& myInterface_)
        : myInterface(myInterface_)
    {
    }

    int getAnInt(int i, int j, int k) const
    {
        return myInterface.intMethod(i, j, k);
    }

    const MyStruct& retrieveStructRef() const
    {
        return myInterface.refStructMethod();
    }

    MyStruct* createStructOnHeap()
    {
        MyStruct* data = myInterface.newStructMethod();

        if (data==NULL)
        {
            throw std::exception("Struct creation failure");
        }

        return data;
    }

private:
    MyInterface& myInterface;
};

class BasicActionFixture : public Test
{
public:
    BasicActionFixture()
        : underTest(mock)
    {
    }

protected:
    MyMock mock;
    UnderTest underTest;
};

TEST_F(BasicActionFixture, shouldReturnSpecifiedValue)
{
    EXPECT_CALL(mock, intMethod(_, _, _)).WillOnce(Return(42));

    EXPECT_EQ(42, underTest.getAnInt(10, 20, 30));
}

TEST_F(BasicActionFixture, shouldReturnSpecifiedArgument)
{
    EXPECT_CALL(mock, intMethod(_, _, _)).WillOnce(ReturnArg<1>());

    EXPECT_EQ(20, underTest.getAnInt(10, 20, 30));
}

TEST_F(BasicActionFixture, shouldReturnNewInstance)
{
    EXPECT_CALL(mock, newStructMethod()).WillOnce(ReturnNew<MyStruct>("Life", 42));

    MyStruct* data = underTest.createStructOnHeap();
    EXPECT_EQ(42, data->count);

    delete data;
}

TEST_F(BasicActionFixture, shouldReturnReferenceToExistingInstance)
{
    MyStruct data("Life", 43);

    EXPECT_CALL(mock, refStructMethod()).WillOnce(ReturnRef(data));

    const MyStruct& dataRef = underTest.retrieveStructRef();
    
    EXPECT_EQ(43, data.count);
    EXPECT_EQ(&data, &dataRef);
}

TEST_F(BasicActionFixture, shouldThrowIfCreateFails)
{
    EXPECT_CALL(mock, newStructMethod()).WillOnce(ReturnNull());

    EXPECT_ANY_THROW(underTest.createStructOnHeap());
}

