#include "gmock/gmock.h"

namespace
{

using namespace testing;

class MyDataObject
{
public:
    virtual ~MyDataObject()
    {
        destruct();
    }

    MOCK_METHOD0(destruct, void());
};

enum Complexity
{
    Easy, QuiteHard, VeryHard, GargantuanMegaBrain
};

class MyInterface
{
public:
    virtual ~MyInterface() {}
    
    virtual void voidMethod() const=0;

    virtual void free(MyDataObject*) =0;

    virtual void answerQuestion(int) =0;

    virtual bool isAnswerReady() const =0;

    virtual int yearsToWait(Complexity) const =0;
};

class MyMock : public MyInterface
{
public:
    MOCK_CONST_METHOD0(voidMethod, void());

    MOCK_METHOD1(free, void(MyDataObject*));

    MOCK_METHOD1(answerQuestion, void(int));

    MOCK_CONST_METHOD0(isAnswerReady, bool());

    MOCK_CONST_METHOD1(yearsToWait, int(Complexity));
};

class UnderTest
{
public:
    UnderTest(MyInterface& myInterface_)
        : myInterface(myInterface_)
    {
    }

    void doSomething()
    {
       myInterface.voidMethod();
    }

    void process(MyDataObject* object)
    {
        myInterface.free(object);
    }

    void initialiseSystem()
    {
        myInterface.answerQuestion(42);
    }

    bool isProcessingComplete() const
    {
        return myInterface.isAnswerReady();
    }

    int generateCogitationEstimate(Complexity complexity) const
    {
        return myInterface.yearsToWait(complexity);
    }

private:
    MyInterface& myInterface;
};

class SideEffectFixture : public Test
{
public:
    SideEffectFixture()
        : underTest(mock)
    {
    }

protected:
    MyMock mock;
    UnderTest underTest;
};

TEST_F(SideEffectFixture, shouldAssignData)
{
    int i = 0;
    EXPECT_CALL(mock, voidMethod()).WillOnce(Assign(&i, 1));

    EXPECT_NO_THROW(underTest.doSomething());
    EXPECT_EQ(1, i);
}

TEST_F(SideEffectFixture, shouldDeleteObject)
{
    MyDataObject* object = new MyDataObject();
    EXPECT_CALL(*object, destruct());

    EXPECT_CALL(mock, free(_)).WillOnce(DeleteArg<0>());

    EXPECT_NO_THROW(underTest.process(object));
}

TEST_F(SideEffectFixture, shouldSaveMeaningOfLife)
{
    int meaning = 0;

    EXPECT_CALL(mock, answerQuestion(_)).WillOnce(SaveArg<0>(&meaning));

    EXPECT_NO_THROW(underTest.initialiseSystem());
    EXPECT_EQ(42, meaning);
}
// Also SetArgReferee<N>, SetArgumentPointee<N>, SetArrayArgument<N>, SetErrnoAndReturn

TEST_F(SideEffectFixture, shouldBeAbleToDoDefaultAction)
{
    EXPECT_CALL(mock, isAnswerReady())
        .WillOnce(DoDefault())
        .WillRepeatedly(Return(true));

    EXPECT_FALSE(underTest.isProcessingComplete());
    EXPECT_TRUE(underTest.isProcessingComplete());
    EXPECT_TRUE(underTest.isProcessingComplete());
}

TEST_F(SideEffectFixture, shouldBeAbleToControlDefaultAction)
{
    ON_CALL(mock, yearsToWait(_)).WillByDefault(Return(26000000));
    ON_CALL(mock, yearsToWait(Easy)).WillByDefault(Return(1));
    ON_CALL(mock, yearsToWait(QuiteHard)).WillByDefault(Return(100));
    ON_CALL(mock, yearsToWait(VeryHard)).WillByDefault(Return(999));

    EXPECT_CALL(mock, yearsToWait(_))
        .WillRepeatedly(DoDefault());

    EXPECT_EQ(26000000, underTest.generateCogitationEstimate(GargantuanMegaBrain));
    EXPECT_EQ(100, underTest.generateCogitationEstimate(QuiteHard));
    EXPECT_EQ(1, underTest.generateCogitationEstimate(Easy));
    EXPECT_EQ(26000000, underTest.generateCogitationEstimate(GargantuanMegaBrain));
}

}