#include "gmock/gmock.h"

namespace
{

using namespace testing;

class MyInterface
{
public:
    virtual ~MyInterface() {}

    virtual int determineBoredomQuotient(int elapsed, int length) const =0;
};

class MyMock : public MyInterface
{
public:
    MOCK_CONST_METHOD2(determineBoredomQuotient, int(int, int));
};

class UnderTest
{
public:
    UnderTest(MyInterface& myInterface_, int sessionLength_)
        : myInterface(myInterface_), boredomThreshold(100), elapsedMinutes(0), sessionLength(sessionLength_)
    {
    }

    bool areWeBoredYet() const
    {
        return myInterface.determineBoredomQuotient(elapsedMinutes, sessionLength) > boredomThreshold;
    }

    void elapse(int time)
    {
        elapsedMinutes += time;
    }

private:
    MyInterface& myInterface;
    int boredomThreshold;
    int elapsedMinutes;
    int sessionLength;
};

class InvocationFixture : public Test
{
public:
    InvocationFixture()
        : underTest(mock, 90)
    {
    }

protected:
    MyMock mock;
    UnderTest underTest;
};

int dummyBoredomGenerator()
{
    return 99;
}

int timeDependentBoredomGenerator(int elapsedTime, int sessionLength)
{
    if (sessionLength - elapsedTime < 10)
    {
        return elapsedTime;
    }
    else
    {
        return elapsedTime * 2;
    }
}


int sessionDependentBoredomGenerator(int sessionLength)
{
    if (sessionLength > 45)
    {
        return 100 + sessionLength;
    }

    return sessionLength;
}

TEST_F(InvocationFixture, dummyShouldNeverBeBored)
{
    EXPECT_CALL(mock, determineBoredomQuotient(_,_)).WillOnce(InvokeWithoutArgs(dummyBoredomGenerator));

    EXPECT_EQ(false, underTest.areWeBoredYet());
}

TEST_F(InvocationFixture, shouldNotBeBoredAtBeginningOfSession)
{
    EXPECT_CALL(mock, determineBoredomQuotient(_,_)).WillOnce(Invoke(timeDependentBoredomGenerator));

    EXPECT_FALSE(underTest.areWeBoredYet());
}

TEST_F(InvocationFixture, shouldNotBeBoredAfterFiftyMinutes)
{
    EXPECT_CALL(mock, determineBoredomQuotient(_,_)).WillOnce(Invoke(timeDependentBoredomGenerator));

    underTest.elapse(50);
    EXPECT_FALSE(underTest.areWeBoredYet());
}

TEST_F(InvocationFixture, shouldBeBoredAfterFiftyOneMinutes)
{
    EXPECT_CALL(mock, determineBoredomQuotient(_,_)).WillOnce(Invoke(timeDependentBoredomGenerator));

    underTest.elapse(51);
    EXPECT_TRUE(underTest.areWeBoredYet());
}

TEST_F(InvocationFixture, shouldBeLookingForwardToCoffee)
{
    EXPECT_CALL(mock, determineBoredomQuotient(_,_)).WillOnce(Invoke(timeDependentBoredomGenerator));

    underTest.elapse(81);
    EXPECT_FALSE(underTest.areWeBoredYet());
}


TEST_F(InvocationFixture, shouldFindAllLongSessionsBoring)
{
    EXPECT_CALL(mock, determineBoredomQuotient(_,_)).WillOnce(WithArg<1>(Invoke(sessionDependentBoredomGenerator)));

    EXPECT_TRUE(underTest.areWeBoredYet());

}

// Also IgnoreResult, DoAll


// User defined actions
ACTION(CalculateBoredomQuotient)
{
    if (arg1 - arg0 < 10)
    {
        return arg0;
    }
    else
    {
        return arg0 * 2;
    }
}

TEST_F(InvocationFixture, shouldNotBeBoredAtBeginningOfSessionUsingAction)
{
    EXPECT_CALL(mock, determineBoredomQuotient(_,_)).WillOnce(CalculateBoredomQuotient());

    EXPECT_FALSE(underTest.areWeBoredYet());
}

TEST_F(InvocationFixture, shouldNotBeBoredAfterFiftyMinutesUsingAction)
{
    EXPECT_CALL(mock, determineBoredomQuotient(_,_)).WillOnce(CalculateBoredomQuotient());

    underTest.elapse(50);
    EXPECT_FALSE(underTest.areWeBoredYet());
}

TEST_F(InvocationFixture, shouldBeBoredAfterFiftyOneMinutesUsingAction)
{
    EXPECT_CALL(mock, determineBoredomQuotient(_,_)).WillOnce(CalculateBoredomQuotient());

    underTest.elapse(51);
    EXPECT_TRUE(underTest.areWeBoredYet());
}

TEST_F(InvocationFixture, shouldBeLookingForwardToCoffeeUsingAction)
{
    EXPECT_CALL(mock, determineBoredomQuotient(_,_)).WillOnce(CalculateBoredomQuotient());

    underTest.elapse(81);
    EXPECT_FALSE(underTest.areWeBoredYet());
}

// Parameterised user defined actions
ACTION_P(CalculateAdjustedBoredomQuotient, adjustment)
{
    if (arg1 - arg0 < 10)
    {
        return arg0 + adjustment;
    }
    else
    {
        return arg0 * 2 + adjustment;
    }
}

TEST_F(InvocationFixture, evenCoffeeCantSaveUs)
{
    EXPECT_CALL(mock, determineBoredomQuotient(_,_)).WillOnce(CalculateAdjustedBoredomQuotient(20));

    underTest.elapse(81);
    EXPECT_TRUE(underTest.areWeBoredYet());
}

// Also ACTION_Pn

}